package de.hawhamburg.reachability.sensor.adapter;


import org.apache.log4j.Logger;

import com.google.gson.Gson;

import de.hawhamburg.reachability.reasoner.enumeration.ReasonerDimension;
import de.hawhamburg.reachability.sensor.Sensor;
import de.hawhamburg.reachability.sensor.SensorAdapter;
import de.hawhamburg.reachability.sensor.SensorAdapterFactory;
import de.hawhamburg.reachability.sensor.SensorAgent;
import de.hawhamburg.reachability.sensor.SensorData;
import de.hawhamburg.reachability.sensor.enumeration.ActivationStates;
import de.hawhamburg.reachability.sensor.enumeration.BedStates;
import de.hawhamburg.reachability.sensor.enumeration.SystemStates;
import de.hawhamburg.reachability.sensor.enumeration.DoorStates;
import de.hawhamburg.reachability.sensor.enumeration.LoadingStates;
import de.hawhamburg.reachability.sensor.enumeration.LocationStates;
import de.hawhamburg.reachability.sensor.enumeration.TVStates;
import de.hawhamburg.reachability.tools.feeder.json.JsonSensorDataFeederItem;
import de.hawhamburg.reachability.util.ConnectionUtil;
import de.hawhamburg.reachability.util.OutputUtil;

/**
 * Receiving simulated sensor data for certain context dimensions.
 * 
 * Objects of this class are providing sensor adapters for simulated sensor data. A {@link RSDFAdapter} will listen
 * on a certain result topic, specified by the XML configuration file in "resource/sensor/RSDFAdapter.xml", for 
 * simulated data for its context dimension. If sensed, the data will be transformed into the internal representation
 * and forwarded to the reasoning agent via the ActiveMQ. 
 * 
 * <b>CAUTION: SIMULATED SENSOR INPUT FROM 
 * {@link de.hawhamburg.reachability.tools.feeder.SensorDataFeeder}</b>
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class RSDFAdapter extends    Sensor
                         implements SensorAdapter  {
        
    /**
     * Creates a new {@link RSDFAdapter} which senses <b>simulated</b>
     * simulated sensor data from
     * {@link de.hawhamburg.reachability.tools.feeder.SensorDataFeeder}.<br><br>
     * 
     * The sensors configurations will be loaded from the XML file in  
     * {@link de.hawhamburg.reachability.sensor.Settings#SENSOR_CONFIGURATION_PATH}
     * which is named like this class.
     * 
     * @param dimension     The {@link ReasonerDimension} for which this sensor is configured.
     */
    public RSDFAdapter(final ReasonerDimension dimension) {
        super(dimension);
    }

    /**
     * Retrieves the fictional stress level of an inhabitant via ActiveMQ and returns it as 
     * {@link SensorData}.
     * 
     * @return              The fictional stress level as {@link SensorData} object. 
     */
    public final SensorData getSensorOutput() {
        String json = ConnectionUtil.getMessageOnTopic(this.getResultTopic());
        
        if (json != null) {
            Gson g = new Gson();
            JsonSensorDataFeederItem msgObj = g.fromJson(json, JsonSensorDataFeederItem.class);
            
            if (ReasonerDimension.valueOf(msgObj.SensorType).equals(this.getDataDimension())) {
                String type  = msgObj.SensorType;
                
                double value = 0.0;
                switch (ReasonerDimension.valueOf(type)) {
                    case location:
                        value = (double) LocationStates.valueOf(msgObj.SensorValue).ordinal();
                        break;

                    case frontdoor:
                    case fridgeDoor:
                    case kitchenCabinet:
                        value = (double) DoorStates.valueOf(msgObj.SensorValue).ordinal();
                        break;
                        
                    case frontdoorBell:
                    case coffeeMachine:
                    case kitchenSink:
                    case fireside:
                    case toiletFlushing:
                    case shower:
                    case toothbrush:
                    case bathTap:
                    case buzzer:
                    case music:
                    case bedroomLight:
                    case toaster:
                        value = (double) ActivationStates.valueOf(msgObj.SensorValue).ordinal();
                        break;
                        
                    case chairDining:
                    case couchLiving:
                    case lavatorySeat:
                        value = (double) LoadingStates.valueOf(msgObj.SensorValue).ordinal();
                        break;
                       
                    case couchTable:
                    case laptop:
                        value = (double) SystemStates.valueOf(msgObj.SensorValue).ordinal();
                        break;
                        
                    case bed:
                        value = (double) BedStates.valueOf(msgObj.SensorValue).ordinal();
                        break;

                    case tvLiving:
                        value = (double) TVStates.valueOf(msgObj.SensorValue).ordinal();
                        break;
                        
                    default:
                        value = Double.parseDouble(msgObj.SensorValue);
                        break;
                }
                
                if (!ReasonerDimension.valueOf(type).equals(ReasonerDimension.time)) {
                    OutputUtil.debug("Received type: " + type + " with value: " + msgObj.SensorValue
                                     + " (Interpreted as: " + value + ")");
                }
                
                return new SensorData(this.getClass().getSimpleName(), this.getDataDimension(), value);
            } else {
                // Wrong data found...
                return null;
            }
        } else {
            // No data found...
            return null;
        }
    }

    /**
     * Main method which starts a single sensor adapter for each context dimension to sense simulated sensor data.
     * 
     * @param args          The start arguments as array of {@link String}.
     */
    public static void main(final String[] args) {
        // Remove all existing loggers
        Logger.getRootLogger().removeAllAppenders();
        
        // Start this sensor adapter
        for (ReasonerDimension dimension : ReasonerDimension.values()) {
            
            // Build this sensor adapter by using the SensorAdapterFactory
            SensorAdapter sa = SensorAdapterFactory.getConnectedSensorAdapter(
                                    RSDFAdapter.class.getSimpleName(),
                                    dimension);
            
            // Start the sensor adapter as sensor agent, which supplied the required control functions
            new SensorAgent(sa);
        }
    }
}
