package de.hawhamburg.reachability.properties;

import java.util.Map;

/**
 * Data object which holds the description of a property.<br><br>
 * 
 * This data objects provides the description of a property which shall be used in common with interface implementation
 * and the configuration management user interface.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class PropertyDescription {

    /**
     * The properties name as {@link String} which shall be displayed within a user interface as name for this property.
     */
    private String displayName;
    
    /**
     * The {@link DisplayType} of this property.
     */
    private DisplayType displayType;
    
    /**
     * The groups name as {@link String} which subsumes this property with others.
     */
    private String displayGroup;
    
    /**
     * The possible values as {@link Map} of {@link String} (value) to {@link String} (display name).
     */
    private Map<String, String> possibleValues;
    
    /**
     * Create empty property description.
     */
    public PropertyDescription() { }
    
    /**
     * Creates a proper property description.
     * 
     * @param dspName       The properties display name as {@link String}.
     * @param dspType       The {@link DisplayType} of this property.
     * @param dspGroup      The groups display name as {@link String} in which this property is included.
     * @param validValues   A list of valid values as {@link Map} of {@link String} (value) to {@link String} 
     *                      (display name).
     */
    public PropertyDescription(final String dspName,
                               final DisplayType dspType,
                               final String dspGroup,
                               final Map<String, String> validValues) {
        this.displayName    = dspName;
        this.displayType    = dspType;
        this.displayGroup   = dspGroup;
        this.possibleValues = validValues;
    }
    
    /**
     * Sets/updates the display name.
     * 
     * @param dspName       The new value for the display name as {@link String}.  
     */
    public final void setDisplayName(final String dspName) {
        this.displayName = dspName;
    }
    
    /**
     * Sets/updates the display type.
     * 
     * @param dspType       The new value for the display type as {@link DisplayType}. 
     */
    public final void setDisplayType(final DisplayType dspType) {
        this.displayType = dspType;
    }
    
    /**
     * Sets/updates the display group.
     * 
     * @param dspGroup      The new value for the display group as {@link String}. 
     */
    public final void setDisplayGroup(final String dspGroup) {
        this.displayGroup = dspGroup;
    }
    
    /**
     * Sets/updates the list of possible values.
     * 
     * @param validValues   The new value for the list of possible values as {@link Map} of {@link String} (value) to
     *                       {@link String} (display name). 
     */
    public final void setPossibleValues(final Map<String, String> validValues) {
        this.possibleValues = validValues;
    }

    /**
     * Adds an entry to the list of possible values.
     * 
     * @param value         The value as {@link String}.
     * @param dspValue      The display name of this value as {@link String} which shall be displayed to the user.
     */
    public final void addPossibleValue(final String value,
                                       final String dspValue) {
        this.possibleValues.put(value, dspValue);
    }
    
    /**
     * Returns the display name as {@link String}. 
     * 
     * @return              The display name as {@link String}.
     */
    public final String getDisplayName() {
        return this.displayName;
    }
    
    /**
     * Returns the display type as {@link DisplayType}. 
     * 
     * @return              The display type as {@link DisplayType}.
     */
    public final DisplayType getDisplayType() {
        return this.displayType;
    }
    
    /**
     * Returns the display group as {@link String}. 
     * 
     * @return              The display group as {@link String}.
     */
    public final String getDisplayGroup() {
        return this.displayGroup;
    }
    
    /**
     * Returns the list of possible values as {@link Map} of {@link String} (value) to {@link String} (display name). 
     * 
     * @return              The list of possible values as {@link Map} of {@link String} (value) to {@link String} 
     *                      (display name).
     */
    public final Map<String, String> getPossibleValue() {
        return this.possibleValues;
    }
}
