package de.hawhamburg.reachability.sensor;

import java.util.HashMap;
import java.util.Map;
import java.util.Random;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPPublisher;
import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPSubscriber;
import de.hawhamburg.reachability.ProjectInfo;
import de.hawhamburg.reachability.StopableObject;
import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.json.JsonReachabilitySensorData;
import de.hawhamburg.reachability.reasoner.enumeration.ReasonerDimension;
import de.hawhamburg.reachability.registry.enumeration.RegistryAgentType;
import de.hawhamburg.reachability.registry.enumeration.RegistryMessageType;
import de.hawhamburg.reachability.registry.json.JsonRARegistryMessage;
import de.hawhamburg.reachability.util.ConnectionUtil;
import de.hawhamburg.reachability.util.OutputUtil;
import de.hawhamburg.reachability.util.XMLUtil;

/**
 * Abstract class which needs to be extended by each {@link SensorAdapter}.<br><br>
 * 
 * This abstract class needs to be extended by a {@link SensorAdapter} to be used in common with the reachability 
 * project. It provides some ActiveMQ connection relevant methods, to extremely simplify the sensor adapter design.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public abstract class Sensor extends    Thread
                             implements StopableObject {
    
    /**
     * Holds the activity state of this object. If silent is <code>true</code>, the worker won't publish its results on 
     * the result topic.
     */
    private boolean silent                          = true;
         
    /**
     * The ActiveMQ topic name on which a request shall be send to the sensors data source.
     */
    private String requestTopicName                 = null;
    
    /**
     * The ActiveMQ topic name on which the sensors data source provides its measured results. 
     */
    private String resultTopicName                  = null;
    
    /**
     * The ActiveMQ topic name on which the {@link Sensor} publishes its generalized data set for the reasoner.
     */
    private String internalSensorDataTopicName      = null;
    
    /**
     * The ActiveMQ topic name on which reachability systems registry service will send its registry invitations.
     */
    private String registryRequestTopicName         = null;
    
    /**
     * The ActiveMQ topic name on which clients send their registration messages to the systems registry service.
     */
    private String registryTopicName                = null;
    
    /**
     * The {@link LPPublisher}, which publishes requests to the sensors data source request topic 
     * {@link #requestTopicName}.
     */
    private LPPublisher requestTopic                = null;
    
    /**
     * The {@link LPSubscriber}, which receives results from the sensors data source result topic 
     * {@link #resultTopicName}.
     */
    private LPSubscriber resultTopic                = null;
    
    /**
     * The {@link LPPublisher}, which publishes the generalized result set on the internal result topic 
     * {@link #internalSensorDataTopicName}.
     */
    private LPPublisher internalSensorDataTopic     = null;
    
    /**
     * The {@link LPSubscriber}, which listens for register requests on the systems registry request topic
     * {@link #registryRequestTopicName}.
     */
    private LPSubscriber requestRegistryTopic       = null;
    
    /**
     * The {@link LPPublisher}, which publishes register messages onto the systems registry register topic
     * {@link #registryTopicName}.
     */
    private LPPublisher registryTopic               = null;
    
    /**
     * The context dimension as {@link ReasonerDimension} for which this sensor supplies data inputs.
     */
    private ReasonerDimension internalDimension     = null;

    /**
     * Determines if the method {@link #shutDown()} has been called.
     */
    private boolean shutDown                        = false;
    
    /**
     * Default delay time.
     */
    private long delay                              = 0;
    
    /**
     * Local random number generator.
     */
    private Random rand                             = null;
    
    /**
     * Local adapter configuration.
     */
    private Map<String, String> localConf           = new HashMap<String, String>();
    
    /**
     * Determines if the {@link Sensor} has stopped.
     */
    private boolean stopped                         = false;
    
    /**
     * Protected constructor ensures, that there will be no direct instance of this class.
     * 
     * @param dimension     The {@link ReasonerDimension} for which this sensor is configured.
     */
    protected Sensor(final ReasonerDimension dimension) {
        this.internalDimension              = dimension;
        this.rand                           = new Random();
        this.internalSensorDataTopicName    = MainConfig.get("INTERNAL_SENSOR_DATA_TOPIC");
        this.registryRequestTopicName       = MainConfig.get("REACHABILITY_REGISTRY_REQUEST_TOPIC");
        this.registryTopicName              = MainConfig.get("REACHABILITY_REGISTRY_TOPIC");
        this.loadDefaultsFromXMLResource();
    }
    
    /**
     * Connects the {@link Sensor} to the topics via ActiveMQ, using the utility class {@link ConnectionUtil}.
     */
    public final void connect() {
        this.requestTopic               = ConnectionUtil.getPublisher(this.requestTopicName, this);
        this.resultTopic                = ConnectionUtil.getSubscriber(this.resultTopicName, this);
        this.internalSensorDataTopic    = ConnectionUtil.getPublisher(this.internalSensorDataTopicName, this);
        this.requestRegistryTopic       = ConnectionUtil.getSubscriber(this.registryRequestTopicName, this);
        this.registryTopic              = ConnectionUtil.getPublisher(this.registryTopicName, this);
        
        String asking = "";
        if (this.requestTopic != null) {
            asking = ", asking on topic '" + this.requestTopicName + "'";
        }
        String listening = "";
        if (this.resultTopic != null) {
            listening = ", listening on topic '" + this.resultTopicName + "'";
        }
        
        OutputUtil.info("Sensor adapter started for data dimension '" + internalDimension + "'" + asking + listening);
    }
    
    /**
     * Sets the silent state. If silent is set to <code>true</code>, the {@link Sensor} won't publish his informations 
     * onto the internal result topic for the reasoner.
     * 
     * @param state         The current state of silence as {@link Boolean}.
     */
    public final void setSilent(final boolean state) {
        this.silent = state;
    }
    
    /**
     * Determines if the {@link Sensor}s running mode is silent.
     * 
     * @return              The current state of silence as {@link Boolean}.
     */
    protected final boolean isSilent() {
        return this.silent;
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#isActive()
     * 
     * @return              <code>true</code> if {@link #shutDown()} wasn't called, otherwise <code>false</code>.
     */
    public final boolean isActive() {
        return !this.shutDown;
    }
    
    /**
     * Determines if the {@link Sensor} has stopped. 
     * 
     * @return              <code>true</code> if {@link Sensor} has stopped, otherwise <code>false</code>.
     */
    public final boolean stopped() {
        return this.stopped;
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#shutDown()
     */
    public final void shutDown() {
        this.shutDown = true;
    }

    /**
     * Provides the name of the request topic for ActiveMQ communication, if this {@link Sensor} is connect to its data
     * source via ActiveMQ.
     * 
     * @return              The request topic name of this {@link Sensor}s data source if its connected via ActiveMQ, 
     *                      otherwise <code>null</code>.
     */
    public final String getRequestTopicName() {
        return this.requestTopicName;
    }

    /**
     * Provides the name of the result topic for ActiveMQ communication, if this {@link Sensor} is connect to its data
     * source via ActiveMQ.
     * 
     * @return              The result topic name of this {@link Sensor}s data source if its connected via ActiveMQ,
     *                      otherwise <code>null</code>.
     */
    public final String getResultTopicName() {
        return this.resultTopicName;
    }
    
    /**
     * Provides the context dimension of the {@link Sensor}s data as {@link ReasonerDimension}.
     * 
     * @return              The {@link ReasonerDimension} of the {@link Sensor}.
     */
    public final ReasonerDimension getDataDimension() {
        return this.internalDimension;
    }
    
    /**
     * Provides the current {@link SensorData} and <code>null</code> if none exists.
     * 
     * @return              A {@link SensorData} object on success, otherwise <code>null</code>.
     */
    public abstract SensorData getSensorOutput();
    
    /**
     * Main run method. Initializes processing by using {@link #sendInitialRegisterMessage()} and frequently does:
     * <ul>
     *  <li>Check for register invitations from the systems registry service by using 
     *      {@link #handleRegisterRequestMessage()}</li>
     *  <li>Check for new sensor data from the {@link Sensor}s data source by using {@link #getSensorOutput()}</li>
     *  <li>Push new sensor data onto the internal data exchange topic by using 
     *      {@link #sendInternalSensorData(SensorData)}</li>
     * </ul>
     * 
     * while {@link #isActive()}. Afterwards cleaning up all resources by using {@link #end()}. 
     */
    public final void run() {
        sendInitialRegisterMessage();
        while (this.isActive()) {
            // Look for register request
            handleRegisterRequestMessage();
            
            // Get sensor data and publish them
            SensorData so = this.getSensorOutput();
            if (so != null) {
                sendInternalSensorData(so);
                try {
                    Thread.sleep(this.delay);
                } catch (InterruptedException e) {
                    OutputUtil.error("[InterruptionExcepion] " + e.getMessage());
                }
            }
        }
        end();
    }
    
    /**
     * Unregisters from systems registry service, disconnects all ActiveMQ topic connections and sets {@link #stopped}
     * to <code>true</code>. 
     */
    private void end() {
        sendRegistryMessage(RegistryMessageType.UNREGISTER);
        
        if (requestTopic != null) {
            requestTopic.disconnect();
        }
        
        if (resultTopic != null) {
            resultTopic.disconnect();
        }
        
        if (internalSensorDataTopic != null) {
            internalSensorDataTopic.disconnect();
        }
        
        if (requestRegistryTopic != null) {
            requestRegistryTopic.disconnect();
        }
        
        if (registryTopic != null) {
            registryTopic.disconnect();
        }

        this.stopped = true;
        
        OutputUtil.info("This sensor for data dimension '" + internalDimension + "' has been gracefully shut down");
    }
    
    /**
     * Sends an initial register message by using {@link #sendRegistryMessage(RegistryMessageType)} with 
     * {@link RegistryMessageType#REGISTER}.
     */
    private void sendInitialRegisterMessage() {
        sendRegistryMessage(RegistryMessageType.REGISTER);
    }
    
    /**
     * Checks for new registry invitations from the systems registry service and sends register message if there's 
     * some by using {@link #sendRegistryMessage(RegistryMessageType)} with {@link RegistryMessageType#REGISTER}.
     */
    private void handleRegisterRequestMessage() {
        String json = ConnectionUtil.getMessageOnTopic(requestRegistryTopic);
        if (json != null) {
            OutputUtil.info("  Recived register request from registry...");
            sendRegistryMessage(RegistryMessageType.REGISTER);
        }
    }
    
    /**
     * Creates and sends a new {@link JsonRARegistryMessage} onto the {@link #registryTopic}. This may register or
     * unregister this {@link Sensor} due to the supplied {@link RegistryMessageType}.
     * 
     * @param type          The {@link RegistryMessageType} which indicates if the {@link Sensor} shall be registered or
     *                      unregistered from the systems registry service.
     */
    private void sendRegistryMessage(final RegistryMessageType type) {
        JsonRARegistryMessage register = new JsonRARegistryMessage();
        
        register.Id             = this.getID();
        register.Version        = this.getVersion();
        register.MessageType    = type;
        register.AgentType      = RegistryAgentType.SENSOR;
        register.Name           = this.getClass().getSimpleName();
        register.Dimension      = this.internalDimension;
        register.Remark         = "Here am I...";
        
        ConnectionUtil.setMessageOnTopic(registryTopic, register);
        
        if (type.equals(RegistryMessageType.REGISTER)) {
            OutputUtil.info("+ Sending new SENSOR registration: '"
                            + register.Name + "' for dimension '" + register.Dimension + "'");
        } else {
            OutputUtil.info("- Sending SENSOR unregistration: '"
                            + register.Name + "' for dimension '" + register.Dimension + "'");
        }
    }
    
    /**
     * Updates or sets the internal request topic name to a new value.
     * 
     * @param newTopicName  The new name of the request topic as {@link String}.
     */
    protected final void setRequestTopicName(final String newTopicName) {
        if ("".equals(newTopicName)) {
            this.requestTopicName = null;
        } else {
            this.requestTopicName = newTopicName;
        }
    }
    
    /**
     * Updates or sets the internal result topic name to a new value.
     * 
     * @param newTopicName  The new name of the result topic as {@link String}.
     */
    protected final void setResultTopicName(final String newTopicName) {
        if ("".equals(newTopicName)) {
            this.resultTopicName = null;
        } else {
            this.resultTopicName = newTopicName;
        }
    }
    
    /**
     * Updates or sets the internal delay time to a new value.
     * 
     * @param newDelay      The new delay time as <code>long</code> representing milliseconds. 
     */
    protected final void setDelay(final long newDelay) {
        this.delay = newDelay;
    }
    
    /**
     * Set a local adapter configuration parameter <i>key</i> to <i>value</i>.
     * 
     * @param key           The key of the local adapter parameter as {@link String}.
     * @param value         The (new) value of the local adapter parameter as {@link String}.
     */
    protected final void setLocalConf(final String key,
                                      final String value) {
        this.localConf.put(key, value);
    }
    
    /**
     * Returns the value of the supplied local adapter parameter <i>key</i> if exiting, otherwise <code>null</code>.
     * 
     * @param key           The local adapter parameter key as {@link String}.
     * @return              The value if the local adapter parameter key as {@link String} if existing, otherwise
     *                      <code>null</code>.
     */
    protected final String getLocalConf(final String key) {
        if (!this.localConf.containsKey(key)) {
            return null;
        }
        
        return this.localConf.get(key);
    }

    /**
     * Returns the current {@link LPSubscriber} object, which is connected to the internally defined result topic of 
     * this {@link Sensor}s data source.
     * 
     * @return              The {@link LPSubscriber} object which is connected to the result topic.
     */
    protected final LPSubscriber getResultTopic() {
        return this.resultTopic;
    }

    /**
     * Returns the current {@link LPPublisher} object, which is connected to the internally defined request topic of 
     * this {@link Sensor}s data source.
     * 
     * @return              The {@link LPPublisher} object which is connected to the request topic.
     */
    protected final LPPublisher getRequestTopic() {
        return this.requestTopic;
    }
    
    /**
     * Loads the sensors local configuration from its XML configuration file. The XML file shall be located at the
     * path defined by {@link Settings#SENSOR_CONFIGURATION_PATH} and be named like the adapters simple class name.
     * <br><br> 
     * 
     * These configuration files needs to contain the following xPaths:
     * <ul>
     *  <li>{@link Settings#SENSOR_RESULT_TOPIC_XPATH}</li>
     *  <li>{@link Settings#SENSOR_RESULT_TOPIC_XPATH}</li>
     *  <li>{@link Settings#SENSOR_RESULT_TOPIC_XPATH}</li>
     * </ul>
     * 
     * Optionally it can contain one or more {@link Settings#SENSOR_LOCAL_CONF_XPATH}.
     */
    protected final void loadDefaultsFromXMLResource() {
        String   path = Settings.SENSOR_CONFIGURATION_PATH + this.getClass().getSimpleName() + ".xml";
        Document conf = XMLUtil.loadDocument(path);
        
        if (conf != null) {
            // Load topics
            this.setRequestTopicName(XMLUtil.getTextContent(conf, Settings.SENSOR_REQUEST_TOPIC_XPATH));
            this.setResultTopicName(XMLUtil.getTextContent(conf, Settings.SENSOR_RESULT_TOPIC_XPATH));
        
            // Load delay time
            this.setDelay(XMLUtil.getIntegerContent(conf, Settings.SENSOR_DELAY_TIME_XPATH));
            
            // Load local adapter parameters
            for (Element e : XMLUtil.elementListByXPath(conf, Settings.SENSOR_LOCAL_CONF_XPATH)) {
                if (e.hasAttribute("name")) {
                    this.setLocalConf(e.getAttribute("name"), e.getTextContent());
                }
            }
        } else {
            OutputUtil.error("No configuration file at '" + path + "' found, system will end!");
            shutDown();
        }
    }
    
    /**
     * Sends the measured {@link SensorData} as common internal representation onto the internal data exchange topic 
     * {@link #internalSensorDataTopicName}, if this {@link Sensor} dosn't run in silent mode ({@link #isSilent()}).
     * 
     * @param so            The measured data as {@link SensorData} object.
     */
    private void sendInternalSensorData(final SensorData so) {
        if (!this.isSilent()) {
            JsonReachabilitySensorData data = getDataMessage(so);
            ConnectionUtil.setMessageOnTopic(internalSensorDataTopic, data);
        }
    }
    
    /**
     * Creates a new {@link JsonReachabilitySensorData} message based on a supplied {@link SensorData}.
     * 
     * @param so            The measured data as {@link SensorData}.
     * @return              A {@link JsonReachabilitySensorData} message based on the supplied {@link SensorData}.
     */
    private JsonReachabilitySensorData getDataMessage(final SensorData so) {
        JsonReachabilitySensorData sensorData = new JsonReachabilitySensorData();
        
        sensorData.Version      = this.getVersion();
        sensorData.Id           = this.getID();
        sensorData.Name         = this.getClass().getSimpleName();
        sensorData.Dimension    = this.getDataDimension();
        sensorData.Values       = so.getValues();
        sensorData.Size         = so.getValues().size();
        
        return sensorData;
    }
    
    /**
     * Returns the ID of this {@link Sensor}.
     * 
     * @return              The ID of this {@link Sensor} as {@link String}.
     */
    public final String getID() {
        return ProjectInfo.REACHABILITY_AGENT_NAME + "_Sensor_"
                + this.getDataDimension() + "_"
                + this.getClass().getSimpleName() + "_"
                + this.rand.nextInt(Integer.MAX_VALUE);
    }
    
    /**
     * Returns the version of this {@link Sensor}.
     * 
     * @return              The version of this {@link Sensor} as {@link String}.
     */
    protected final String getVersion() {
        return ProjectInfo.REACHABILITY_AGENT_VERSION;
    }
}
